import 'dart:math';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'dart:ui' as ui;
import 'dart:typed_data';
import 'math_more.dart';
num degToRad(num deg) => deg * (pi / 180.0);

num radToDeg(num rad) => rad * (180.0 / pi);
//double get percent => (progress - min).toDouble() / (max - min);
int toProgress(double percent, {int min = 0, int max = 100}) {
  return (percent * (max - min)).toInt() + min;
}
double calScale(double targetLength, double originLength,double ratio){
  return targetLength * ratio / originLength;
}
Future<ui.Image> load(String asset) async {
  if(asset == null || asset == "") return null;
  ByteData data = await rootBundle.load(asset);
  ui.Codec codec = await ui.instantiateImageCodec(data.buffer.asUint8List());
  ui.FrameInfo fi = await codec.getNextFrame();
  return fi.image;
}

typedef OnProgressChanged<int> = void Function(int value);

class Croller extends StatefulWidget {
  final double radius;
  final double dotRadius;
  final double shadowWidth;
  final Color shadowColor;
  final Color dotColor;
  final Color dotEdgeColor;
  final Color ringColor;
  final OnProgressChanged onProgressChanged;
  final int progress;
  final int min,max;
  final int startAngleOffset;
  final String mainImagePath,backImagePath;
  final double mainRatio,backRatio;
  final bool touchAble;
  const Croller({
    Key key,
    @required this.radius,
    @required this.dotRadius,
    @required this.dotColor,
    this.shadowWidth = 2.0,
    this.shadowColor = Colors.black12,
    this.ringColor = const Color(0XFFF7F7FC),
    this.dotEdgeColor = const Color(0XFFF5F5FA),
    this.progress,
    this.onProgressChanged,
    this.min = 0,
    this.max = 100,
    this.startAngleOffset = 30,
    this.mainImagePath = 'assets/images/knob_main_n.png',
    this.backImagePath = 'assets/images/knob_back_n.png',
    this.mainRatio = -1, this.backRatio = -1,
    this.touchAble = true
  }) : super(key: key);

  @override
  State<StatefulWidget> createState() => _CircleProgressState();
}

class _CircleProgressState extends State<Croller> with SingleTickerProviderStateMixin {
  AnimationController progressController;
  final GlobalKey paintKey = GlobalKey();
  double get backRatio => widget.backRatio == -1 ? 1 : widget.backRatio;
  double get mainRatio => widget.mainRatio == -1 ? 1 : widget.mainRatio;
  double get _width {
    RenderBox box = paintKey.currentContext.findRenderObject();
    return  box.size.width;
  }
  double get _height {
    RenderBox box = paintKey.currentContext.findRenderObject();
    return  box.size.height;
  }
  double currTouchDeg = 0, deg = 0, downTouchDeg = 0;

  final isAntiClockwise = false;
  @override
  void initState() {
    super.initState();
    progressController = AnimationController(duration: Duration(milliseconds: 300), vsync: this);
    if (widget.progress != null) {
      var percent = (widget.progress - widget.min).toDouble() / (widget.max - widget.min);
      progressController.value = percent;
      deg = (progressController.value * (360 - widget.startAngleOffset * 2));
    }

    progressController.addListener(() {
      if (widget.onProgressChanged != null) {
        final progress = toProgress(
          progressController.value,
          min: widget.min,
          max: widget.max,
        );
        widget.onProgressChanged(progress);
      }
      setState(() {});
    });
  }

  @override
  void dispose() {
    progressController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final double width = widget.radius * 2.0;
    final size = new Size(width, width);
    final centre = Offset(size.width / 2, size.height/ 2);
    final angle = (progressController.value * (360 - widget.startAngleOffset * 2)).floor();
    return GestureDetector(
      onPanDown: _onPanDown,
      onPanStart: _onPanStart,
      onPanUpdate: _onPanUpdate,
      onPanEnd: _onPanEnd,
      child:Container(
        width: size.width,
        height: size.height,
        child:  Stack(
          alignment: Alignment.center,
          fit: StackFit.expand,
          children: <Widget>[
            Transform(
              transform: Matrix4.diagonal3Values(backRatio, backRatio, 1),
              origin: centre,
              child: Image.asset(widget.backImagePath,
                fit: BoxFit.contain,
              ),
            ),
            Transform(
              transform: Matrix4.rotationZ(degToRad(angle + 180 + widget.startAngleOffset).toDouble())
                ..scale(mainRatio),
              origin: centre,
              child: Image.asset(widget.mainImagePath,
                fit: BoxFit.contain,
              ),
            ),
            Container(
//              alignment: FractionalOffset.center,
              child: CustomPaint(
                key: paintKey,
                size: size,
                painter: ProgressPainter(
                  dotRadius: widget.dotRadius,
                  shadowWidth: widget.shadowWidth,
                  shadowColor: widget.shadowColor,
                  ringColor: widget.ringColor,
                  dotColor: widget.dotColor,
                  dotEdgeColor: widget.dotEdgeColor,
                  angle: angle.toDouble(),
                  startAngleOffset: widget.startAngleOffset,
                ) ,
              ),
            ),
          ],
        ),
      ),
    );
  }
  void _onPanDown(DragDownDetails details){
    RenderBox getBox = paintKey.currentContext.findRenderObject();
    Offset local = getBox.globalToLocal(details.globalPosition);
    double dx = local.dx - _width / 2;
    double dy = local.dy - _height / 2;
    downTouchDeg = ((atan2(dy, dx) * 180) / pi);
  }
  void _onPanStart(DragStartDetails details) {
    if (!widget.touchAble) {
      return;
    }
    RenderBox getBox = paintKey.currentContext.findRenderObject();
    Offset local = getBox.globalToLocal(details.globalPosition);
  }

  void _onPanUpdate(DragUpdateDetails details) {
    if (!widget.touchAble) {
      return;
    }

    RenderBox getBox = paintKey.currentContext.findRenderObject();
    Offset local = getBox.globalToLocal(details.globalPosition);
    final double center = widget.radius;
    double dx = local.dx - _width / 2;
    double dy = local.dy - _height / 2;
    currTouchDeg = (atan2(dy, dx) * 180) / pi;
    double ddeg = currTouchDeg - downTouchDeg;
    if(abs(ddeg) > 90){
    }else{
      setDeg(deg + ddeg);
    }
    downTouchDeg = currTouchDeg;
  }
  void setDeg(double newDeg) {
    double maxDeg = (360 - widget.startAngleOffset * 2).toDouble();
    deg = newDeg;
    if(newDeg < 0){
      deg = 0;
    }else if (deg > maxDeg){
      deg = maxDeg;
    }
    progressController.value =  newDeg / (360 - widget.startAngleOffset * 2);
  }
  void _onPanEnd(DragEndDetails details) {
    if (!widget.touchAble) {
      return;
    }
  }
}

class ProgressPainter extends CustomPainter {
  final double dotRadius;
  final double shadowWidth;
  final Color shadowColor;
  final Color dotColor;
  final Color dotEdgeColor;
  final Color ringColor;
  final double angle;
  final int startAngleOffset;
  ProgressPainter({
    this.dotRadius,
    this.shadowWidth = 2.0,
    this.shadowColor = Colors.black12,
    this.ringColor = const Color(0XFFF7F7FC),
    this.dotColor,
    this.dotEdgeColor = const Color(0XFFF5F5FA),
    this.angle = 0,
    this.startAngleOffset = 30,
  });

  @override
  void paint(Canvas canvas, Size size) {
    final double centerW = size.width * 0.5;
    final double centerH = size.height * 0.5;
    final center = centerW < centerH ? centerW : centerH;
    final Offset offsetCenter = Offset(centerW, centerH);
    final double drawRadius = size.width * 0.5 - dotRadius;
    final double radians = degToRad(angle);

    final double radiusOffset = dotRadius * 0.4;
    final double outerRadius = center;
    final double innerRadius = center - dotRadius * 2 + radiusOffset;

    final Rect arcRect = Rect.fromCircle(center: offsetCenter, radius: drawRadius);

    // draw ring.
    final ringPaint = Paint()
      ..style = PaintingStyle.stroke
      ..color = ringColor
      ..strokeCap = StrokeCap.round
      ..strokeWidth = (outerRadius - innerRadius);
    canvas.drawArc(arcRect, degToRad(startAngleOffset + 90), degToRad(360 - startAngleOffset * 2), false, ringPaint);

    final Color currentDotColor = Color.alphaBlend(
        dotColor.withOpacity(0.7 + (0.3 * angle / 360)), Colors.white);

    // draw .
    if (angle > 0.0) {
      final progressWidth = outerRadius - innerRadius + radiusOffset;
      canvas.translate(size.width,0);
      canvas.rotate(degToRad(90));
      final Gradient gradient = new SweepGradient(
        endAngle: degToRad(360),
        colors: [
          Colors.white,
          currentDotColor,
        ],
      );
      final progressPaint = Paint()
        ..style = PaintingStyle.stroke
        ..strokeCap = StrokeCap.round
        ..strokeWidth = progressWidth
        ..shader = gradient.createShader(arcRect);
      canvas.drawArc(arcRect, degToRad(startAngleOffset), degToRad(angle), false, progressPaint);
    }
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return true;
  }
}
